@@ -1,69 +1,64 @@
 import asyncio
 from abc import abstractmethod, ABC
-from typing import Union, Type, TypeVar, Iterator, Optional, List, Generic
+from typing import Union, Type, TypeVar, Iterator, Optional, List, Generic, TYPE_CHECKING
 
-from sqlalchemy import and_, select, Table, exists
+from sqlalchemy import and_, select, Table, exists, MetaData
 
 from fox_orm import FoxOrm
-from fox_orm.exceptions import WTFException, NotFetchedException, OrmException
-from fox_orm.utils import full_import, OptionalAwaitable
+from fox_orm.exceptions import NotFetchedException, OrmException
+from fox_orm.internal.utils import full_import, OptionalAwaitable
+
+if TYPE_CHECKING:
+    from fox_orm.model import OrmModel
 
 MODEL = TypeVar('MODEL', bound='OrmModel')
 RELATION = TypeVar('RELATION', bound='_GenericIterableRelation')
 
 
-class IdsList(list):
-    ids: set
+class HashList(List[MODEL]):
+    map: dict
 
     def __init__(self, items: Optional[List[MODEL]] = None):  # pylint: disable=unsubscriptable-object
-        self.ids = set()
+        self.map = dict()
         if items is not None:
             super().__init__(items)
-            for x in items:
-                self.ids.add(x.id)
+            for i, x in enumerate(items):
+                self.map[x.pkey_value] = i
         else:
             super().__init__()
 
     def add(self, other: MODEL):
-        if other.id in self.ids:
+        if other.pkey_value in self.map:
             return
         self.append(other)
-        self.ids.add(other.id)
+        self.map[other.pkey_value] = len(self) - 1
 
     def delete(self, other: MODEL):
-        if other.id not in self.ids:
+        if other.pkey_value not in self.map:
             return
-        for i, x in enumerate(self):
-            if x.id == other.id:
-                to_delete = i
-                break
-        else:
-            raise WTFException('Object not in list')
-        del self[to_delete]
-        self.ids.remove(other.id)
+        del self[self.map[other.pkey_value]]
+        del self.map[other.pkey_value]
 
-    def __contains__(self, item: Union[MODEL, int]):  # pylint: disable=unsubscriptable-object
+    def __contains__(self, item: Optional[Union[MODEL, int]]):  # pylint: disable=unsubscriptable-object
         if item is None:
             return False
         if isinstance(item, int):
-            return item in self.ids
-        return item.id in self.ids
+            return item in self.map
+        return item.pkey_value in self.map
 
-    def __and__(self, other: 'IdsList'):
+    def __and__(self, other: 'HashList'):
         result = []
         for x in other:
-            if x.id in self.ids:
+            if x.pkey_value in self.map:
                 result.append(x)
         return result
 
-    def __or__(self, other: 'IdsList'):
+    def __or__(self, other: 'HashList'):
         result = []
-        ids = set()
         for x in other:
             result.append(x)
-            ids.add(x.id)
         for x in self:
-            if x.id not in ids:
+            if x.pkey_value not in other.map:
                 result.append(x)
         return result
 
@@ -71,19 +66,23 @@ class IdsList(list):
 class _GenericIterableRelation(ABC):
     _fetched: bool
     _initialized: bool
-    _objects: IdsList
+    _objects: HashList
     __modified__: dict
 
-    model: MODEL
+    _model: 'OrmModel'
 
     def __init__(self):
-        self._objects = IdsList()
+        self._objects = HashList()
         self.__modified__ = dict()
         self._fetched = False
         self._initialized = False
 
     @abstractmethod
-    def _init_copy(self: RELATION, model: MODEL) -> RELATION:
+    def _init(self: RELATION, metadata: MetaData, _from: 'Type[OrmModel]'):
+        ...
+
+    @abstractmethod
+    def _init_copy(self: RELATION, model: 'OrmModel') -> RELATION:
         ...
 
     @abstractmethod
@@ -101,14 +100,13 @@ class _GenericIterableRelation(ABC):
 
     async def fetch(self) -> None:
         self._raise_if_not_initialized()
-        self.model.ensure_id()
+        self._model.ensure_id()
         ids = await self.fetch_ids()
-        self._objects = IdsList(await asyncio.gather(*[self.objects_type.get(x) for x in ids]))
+        self._objects = HashList(await asyncio.gather(*[self.objects_type.get(x) for x in ids]))
         self._fetched = True
 
     def _raise_if_not_initialized(self):
-        if not self._initialized:
-            raise WTFException('Relation not initialized')
+        assert self._initialized
 
     def _raise_if_not_fetched(self):
         self._raise_if_not_initialized()
@@ -121,7 +119,7 @@ class _GenericIterableRelation(ABC):
         if not isinstance(other, self.objects_type):
             raise OrmException('other is not instance of target model')
         self._objects.add(other)
-        self.__modified__[other.id] = True
+        self.__modified__[other.pkey_value] = True
         return OptionalAwaitable(self.save)
 
     def delete(self, other: MODEL):
@@ -130,7 +128,7 @@ class _GenericIterableRelation(ABC):
         if not isinstance(other, self.objects_type):
             raise OrmException('other is not instance of target model')
         self._objects.delete(other)
-        self.__modified__[other.id] = False
+        self.__modified__[other.pkey_value] = False
         return OptionalAwaitable(self.save)
 
     def __contains__(self, item: Union[MODEL, int]) -> bool:  # pylint: disable=unsubscriptable-object
@@ -173,57 +171,69 @@ class _GenericIterableRelation(ABC):
 
 
 class ManyToMany(Generic[MODEL], _GenericIterableRelation):
-    to: Union[Type[MODEL], str]  # pylint: disable=unsubscriptable-object
-    via: Table
-    this_id: str
-    other_id: str
+    _from: 'Type[OrmModel]'
+    _to: Union[Type[MODEL], str]  # pylint: disable=unsubscriptable-object
+    _via: Table
+    _via_name: str
+    _this_id: str
+    _other_id: str
 
     async def _get_entry(self, other_id):
         return await FoxOrm.db.fetch_val(select([exists().where(and_(
-            getattr(self.via.c, self.this_id) == self.model.id,
-            getattr(self.via.c, self.other_id) == other_id
+            getattr(self._via.c, self._this_id) == self._model.pkey_value,
+            getattr(self._via.c, self._other_id) == other_id
         ))]))
 
     # pylint: disable=unsubscriptable-object
-    def __init__(self, to: Union[Type[MODEL], str], via: Table, this_id: str, other_id: str):
-        self.to = to
-        self.via = via
-        self.this_id = this_id
-        self.other_id = other_id
+    def __init__(self, to: Union[Type[MODEL], str], via: str):
+        self._to = to
+        self._via_name = via
         super().__init__()
 
-    def _init_copy(self: 'ManyToMany', model: MODEL) -> 'ManyToMany':
-        if isinstance(self.to, str):
-            self.to = full_import(self.to)
-        res = ManyToMany(to=self.to, via=self.via, this_id=self.this_id, other_id=self.other_id)
-        res.model = model
-        res._initialized = True  # pylint: disable=protected-access
+    def _init(self, metadata: MetaData, _from: 'Type[OrmModel]'):
+        self._from = _from
+        if isinstance(self._to, str):
+            self._to = full_import(self._to)
+        self._via, self._this_id, self._other_id = \
+            FoxOrm.get_assoc_table(metadata, self._from, self._to, self._via_name)
+
+    # pylint: disable=protected-access
+    def _init_copy(self: 'ManyToMany', model: 'OrmModel') -> 'ManyToMany':
+        res = ManyToMany(to=self._to, via=self._via)
+        res._model = model
+        res._initialized = True
+        res._via = self._via
+        res._from = self._from
+        res._this_id = self._this_id
+        res._other_id = self._other_id
         return res
 
     @property
     def objects_type(self) -> Type[MODEL]:
         self._raise_if_not_initialized()
-        return self.to
+        return self._to
 
     async def fetch_ids(self) -> List[int]:
-        return [x[self.other_id] for x in await FoxOrm.db.fetch_all(self.via.select().where(
-            getattr(self.via.c, self.this_id) == self.model.id
+        self._raise_if_not_initialized()
+        return [x[self._other_id] for x in await FoxOrm.db.fetch_all(self._via.select().where(
+            getattr(self._via.c, self._this_id) == self._model.pkey_value
         ))]
 
     async def save(self) -> None:
-        self.model.ensure_id()
+        self._model.ensure_id()
+        self._raise_if_not_initialized()
         queries = []
         for k, v in self.__modified__.items():
             entry_exists = await self._get_entry(k)
             if v and not entry_exists:
-                queries.append(FoxOrm.db.execute(self.via.insert(), {
-                    self.this_id: self.model.id,
-                    self.other_id: k
+                queries.append(FoxOrm.db.execute(self._via.insert(), {
+                    self._this_id: self._model.pkey_value,
+                    self._other_id: k
                 }))
             elif not v and entry_exists:
-                queries.append(FoxOrm.db.execute(self.via.delete().where(and_(
-                    getattr(self.via.c, self.this_id) == self.model.id,
-                    getattr(self.via.c, self.other_id) == k
+                queries.append(FoxOrm.db.execute(self._via.delete().where(and_(
+                    getattr(self._via.c, self._this_id) == self._model.pkey_value,
+                    getattr(self._via.c, self._other_id) == k
                 ))))
         await asyncio.gather(*queries)
         self.__modified__ = dict()
@@ -232,11 +242,12 @@ class ManyToMany(Generic[MODEL], _GenericIterableRelation):
 class OneToMany(Generic[MODEL], _GenericIterableRelation):
     to: Union[Type[MODEL], str]  # pylint: disable=unsubscriptable-object
     key: str
+    _from: 'Type[OrmModel]'
 
     async def _get_entry(self, other_id):
         return await FoxOrm.db.fetch_val(select([exists().where(and_(
-            getattr(self.to.c, self.key) == self.model.id,
-            self.to.c.id == other_id
+            getattr(self.to.c, self.key) == self._model.pkey_value,
+            self.to.pkey_column == other_id
         ))]))
 
     # pylint: disable=unsubscriptable-object
@@ -245,11 +256,14 @@ class OneToMany(Generic[MODEL], _GenericIterableRelation):
         self.key = key
         super().__init__()
 
-    def _init_copy(self: 'OneToMany', model: MODEL) -> 'OneToMany':
+    def _init(self, metadata: MetaData, _from: 'Type[OrmModel]'):
+        self._from = _from
         if isinstance(self.to, str):
             self.to = full_import(self.to)
+
+    def _init_copy(self: 'OneToMany', model: MODEL) -> 'OneToMany':
         res = OneToMany(to=self.to, key=self.key)
-        res.model = model
+        res._model = model  # pylint: disable=protected-access
         res._initialized = True  # pylint: disable=protected-access
         return res
 
@@ -259,26 +273,29 @@ class OneToMany(Generic[MODEL], _GenericIterableRelation):
         return self.to
 
     async def fetch_ids(self) -> List[int]:
-        return [x['id'] for x in await FoxOrm.db.fetch_all(select([self.to.c.id]).where(
-            getattr(self.to.c, self.key) == self.model.id
+        return [x[self.to.__pkey_name__] for x in await FoxOrm.db.fetch_all(select([self.to.pkey_column]).where(
+            getattr(self.to.c, self.key) == self._model.pkey_value
         ))]
 
     async def save(self) -> None:
-        self.model.ensure_id()
+        self._model.ensure_id()
         queries = []
         for k, v in self.__modified__.items():
             entry_exists = await self._get_entry(k)
             if v and not entry_exists:
-                queries.append(FoxOrm.db.execute(self.to.__sqla_table__.update().where(
-                    self.to.c.id == k
+                queries.append(FoxOrm.db.execute(self.to.__table__.update().where(
+                    self.to.pkey_column == k
                 ), {
-                    self.key: self.model.id
+                    self.key: self._model.pkey_value
                 }))
             elif not v and entry_exists:
-                queries.append(FoxOrm.db.execute(self.to.__sqla_table__.update().where(
-                    self.to.c.id == k
+                queries.append(FoxOrm.db.execute(self.to.__table__.update().where(
+                    self.to.pkey_column == k
                 ), {
                     self.key: None
                 }))
         await asyncio.gather(*queries)
         self.__modified__ = dict()
+
+
+__all__ = ['ManyToMany', 'OneToMany']
